/*
 * Copyright (c) 2010
 *	HIROSHI OOTA.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <fcntl.h>
#include <sysexits.h>

#include "conf.h"
#include "privs.h"
#define do_PRIVS_ROOT do {PRIVS_ROOT} while(0)
#define do_PRIVS_RELINQUISH do {PRIVS_RELINQUISH} while(0)
#define do_PRIVS_USER do {PRIVS_USER} while(0)

#include "rdns.h"

#define SERVICE_NAME "netbios-ns"

#define INI_READONLY

struct nbt_header {
  u_int16_t nh_xid;
#if BYTE_ORDER == BIG_ENDIAN
  u_int8_t nh_opcode:5;
  u_int8_t nh_nmflagsh:3;
  u_int8_t nh_nmflagsl:4;
  u_int8_t nh_rcode:4;
#endif
#if BYTE_ORDER == LITTLE_ENDIAN
  u_int8_t nh_nmflagsh:3;
  u_int8_t nh_opcode:5;
  u_int8_t nh_rcode:4;
  u_int8_t nh_nmflagsl:4;
#endif
  u_int16_t nh_qdcount, nh_ancount,nh_nscount, nh_arcount;
} __packed;

#define NBT_NAMELENGTH 15
struct nbt_name {
  u_int8_t nn_length;
  u_char nn_name[(NBT_NAMELENGTH + 1) * 2 + 1];
} __packed;

struct nbt_query {
  u_int16_t nq_type;
  u_int16_t nq_class;
} __packed;


struct nbt_response {
  u_int16_t nr_type;
  u_int16_t nr_class;
  u_int32_t nr_ttl;
  u_int16_t nr_length;
  u_int16_t nr_ansflags;
  struct in_addr nr_addr;
} __packed;

struct device_info {
  struct device_info *di_next;
  u_char *di_name;
  struct in_addr di_addr;
  u_char di_devlist;
};

struct group_info {
  struct group_info *gi_next;
  char *gi_name;
  struct device_info *gi_device;
};

static struct addrinfo* gethost(const char *host, const char *port)
{
  struct addrinfo hints, *res;
  int error;

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = PF_INET;
  hints.ai_socktype = SOCK_DGRAM;
  hints.ai_flags = AI_PASSIVE;

  if ((error = getaddrinfo(host, port, &hints, &res))) {
      pr_log_pri(PR_LOG_ERR, "getaddrinfo: %s", gai_strerror(error));
      return NULL;
  }
  return res;
}

static int create_socket(const char *port, int reuse_addr)
{
    struct addrinfo *res, *res0;
    int sock = -1, one = 1, err;

    if ((res0 = gethost(NULL, port)) == NULL)
      return -1;

    for (res = res0; res ; res = res->ai_next) {
	sock = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
	if (sock < 0) {
	    perror("socket");
	    continue;
	}

	if (reuse_addr && 
	    setsockopt(sock, SOL_SOCKET, SO_REUSEPORT, &one, sizeof one) < 0) {
	    perror("setsockopt");
	    close(sock);
	    sock = -1;
	    continue;
	}
    
	do_PRIVS_ROOT;
	err = bind(sock, res->ai_addr, res->ai_addrlen);
	do_PRIVS_RELINQUISH;

	if (err < 0) {
	    perror("bind");
	    close(sock);
	    sock = -1;
	    continue;
	}
	{
	    char buf[INET_ADDRSTRLEN+1];
	    inet_ntop(res->ai_family, 
		      &((struct sockaddr_in *)(res->ai_addr))->sin_addr,
		      buf, sizeof buf);

	}
	break;
    }
    freeaddrinfo(res0);
    return sock;
}

static u_char *getname(struct nbt_name *np, const void *end, 
		       u_int8_t *name_type, u_char *namebuf)
{
    if ((np->nn_length & 0xc0) == 0xc0) {
      np = (struct nbt_name *)
	((u_char *)np + (ntohs(*(unsigned short*)(&np->nn_length)) & 0x3ff));
      if ((void *)np > end) {
	pr_log_pri(PR_LOG_ERR, "corrupt length ? %016x", 
		   ntohs(*(unsigned short*)&np->nn_length));
	return NULL;
      }
    }
    
    int len = np->nn_length / 2;
    int i;

    for (i = 0; i < len; i++) {
      if (np->nn_name[i*2] < 'A' || np->nn_name[i*2] > 'P' ||
	  np->nn_name[i*2+1] < 'A' || np->nn_name[i*2+1] > 'P')
	break;
      namebuf[i] = ((np->nn_name[i*2] - 'A') << 4) + (np->nn_name[i*2+1] - 'A');
    }
    if (name_type)
      *name_type = namebuf[i-1];
    namebuf[i-1] = 0;
    for (i -= 1;i > 0; i--) {
      if (namebuf[i] != ' ')
	break;
      namebuf[i] = 0;
    }
      
    return namebuf;
}

static void response(const char *hostname,
		     int devlist,
		     u_int16_t xid, int sock,
		     struct sockaddr *to, socklen_t tolen)
{
    u_char data[sizeof(struct nbt_header) +
		sizeof(struct nbt_name) +
		sizeof(struct nbt_response)];
    int i;
    struct nbt_header *hp = (struct nbt_header*)data;
    struct nbt_name *np = (struct nbt_name *)(hp + 1);
    struct nbt_response *rp = (struct nbt_response *)(np + 1);
    
    bzero(data, sizeof data);
    
    hp->nh_opcode = 0x10;
    hp->nh_nmflagsh = 0x5;
    hp->nh_ancount = htons(1);
    
    rp->nr_type = htons(0x20);
    rp->nr_class = htons(0x01);
    rp->nr_ttl = 0;
    rp->nr_length = htons(sizeof(struct in_addr) + sizeof(u_int16_t));
    rp->nr_ansflags = 0;
    
    for (i = 0; i < NBT_NAMELENGTH; i++) {
	u_char c =  (*hostname) ? *(hostname++) : ' ';
	np->nn_name[i * 2] = (c >> 4) + 'A';
	np->nn_name[i * 2 + 1] = (c &0xf) + 'A';
    }
    np->nn_length = (i + 1) * 2;
    np->nn_name[i * 2] = np->nn_name[i * 2 + 1] = 0 + 'A'; // name_type
    
    rp->nr_addr.s_addr = INADDR_ANY;

    hp->nh_xid = htons((xid & 0xff) + (devlist << 8));

    if (sendto(sock, data, sizeof(data), 0, to, tolen) < sizeof(data))
      pr_log_pri(PR_LOG_ERR, "sendto: %s", strerror(errno));
}

void *rdns_main(struct rdns_arg *arg)
{
    int sock = arg->sock_fd;
   
    while(1) {
	u_char buf[1024];
	struct sockaddr_in peer;
	int len;
	socklen_t addr_len;
	struct nbt_header *hp = (struct nbt_header *)buf;
	struct nbt_name *np = (struct nbt_name *)((struct nbt_header *)buf + 1);
	
	addr_len = sizeof peer;
	len = recvfrom(sock, buf, sizeof buf, 0, (struct sockaddr*)&peer, &addr_len);

	struct nbt_query *qp =
	    (struct nbt_query *)((u_char *)np + 
				 (((np->nn_length & 0xc0) == 0xc0) ? 
				  2 : np->nn_length + 2));
	if (hp->nh_opcode != 0 ||
	    (hp->nh_nmflagsh << 4) + hp->nh_nmflagsl != 0x11 ||
	    ntohs(hp->nh_qdcount) != 1 ||
	    ntohs(qp->nq_type) != 0x20 || 
	    ntohs(qp->nq_class) != 0x01)
	  continue;
	
	u_char name_type;
	u_char namebuf[sizeof(np->nn_name)];

	u_char *name = getname(np, buf + sizeof(buf), &name_type, namebuf);
	
	if (!name || name_type != 0x00)
	    continue;
	
	if (!strcmp(arg->group_name, (char *)name)) {
	    response(arg->server_name, arg->device_list, ntohs(hp->nh_xid),
		     sock, (struct sockaddr*)&peer, addr_len);
	}
    }

    return NULL;
}

int rdns_create_socket(struct rdns_arg *arg)
{
    if ((arg->sock_fd = create_socket(SERVICE_NAME, arg->reuse_addr)) < 0) 
      return errno;
    return 0;
}
